@@ -11,7 +11,8 @@ Copyright (c) 2000 Fourthought Inc, USA.   All Rights Reserved.
 See  http://4suite.com/COPYRIGHT  for license and copyright information
 """
 
-"""Some Helper functions: 4DOM-specific Extensions to the DOM"""
+"""Some Helper functions: 4DOM/PyXML-specific Extensions to the DOM,
+and DOM-related utilities."""
 
 import sys,string
 
@@ -19,6 +20,7 @@ from xml.dom import Node
 from xml.dom.NodeFilter import NodeFilter
 from xml.dom import XML_NAMESPACE, XMLNS_NAMESPACE, DOMException
 from xml.dom.html import HTML_4_TRANSITIONAL_INLINE
+from c14n import Canonicalize
 import re
 
 
@@ -71,7 +73,8 @@ def PrettyPrint(root, stream=sys.stdout, encoding='UTF-8', indent='  ',
     from xml.dom.ext import Printer
     nss_hints = SeekNss(root)
     preserveElements = preserveElements or []
-    if hasattr(root.ownerDocument, 'getElementsByName'):
+    owner_doc = root.ownerDocument or root
+    if hasattr(owner_doc, 'getElementsByName'):
         #We don't want to insert any whitespace into HTML inline elements
         preserveElements = preserveElements + HTML_4_TRANSITIONAL_INLINE
     visitor = Printer.PrintVisitor(stream, encoding, indent,
@@ -118,7 +121,9 @@ def StripHtml(startNode, preserveElements=None):
     preserveElements = preserveElements or []
     preserveElements = preserveElements + HTML_4_TRANSITIONAL_INLINE
     remove_list = []
-    snit = startNode.ownerDocument.createNodeIterator(startNode, NodeFilter.SHOW_TEXT, None, 0)
+    owner_doc = startNode.ownerDocument or startNode
+    snit = owner_doc.createNodeIterator(startNode, NodeFilter.SHOW_TEXT,
+                                        None, 0)
     curr_node = snit.nextNode()
     while curr_node:
         #first of all make sure it is not inside one of the preserve_elements
@@ -145,7 +150,9 @@ def StripXml(startNode, preserveElements=None):
     '''
     preserveElements = preserveElements or []
     remove_list = []
-    snit = startNode.ownerDocument.createNodeIterator(startNode, NodeFilter.SHOW_TEXT, None, 0)
+    owner_doc = startNode.ownerDocument or startNode
+    snit = owner_doc.createNodeIterator(startNode, NodeFilter.SHOW_TEXT,
+                                        None, 0)
     curr_node = snit.nextNode()
     while curr_node:
         #first of all make sure it is not inside xml:space='preserve'
@@ -166,21 +173,23 @@ def StripXml(startNode, preserveElements=None):
         node_to_remove.parentNode.removeChild(node_to_remove)
     return startNode
 
+_id_key = ('', 'ID')
 
 def GetElementById(startNode, targetId):
     '''
     Return the element in the given tree with an ID attribute of the given
     value
     '''
-    result = None
-    snit = startNode.ownerDocument.createNodeIterator(startNode, NodeFilter.SHOW_ELEMENT, None, 0)
+    owner_doc = startNode.ownerDocument or startNode
+    snit = owner_doc.createNodeIterator(startNode, NodeFilter.SHOW_ELEMENT,
+                                        None, 0)
     curr_node = snit.nextNode()
-    while not result and curr_node:
-        attrs = curr_node.attributes
-        if attrs.has_key("ID") and attrs["ID"] == target_id:
-            result = curr_node
+    while curr_node:
+        attr = curr_node.attributes.get(_id_key, None)
+        if attr and attr._get_nodeValue() == targetId:
+            return curr_node
         curr_node = snit.nextNode()
-    return result
+    return None
 
 
 def XmlSpaceState(node):
@@ -213,7 +222,7 @@ def GetAllNs(node):
         for attr in node.attributes.values():
             if attr.namespaceURI == XMLNS_NAMESPACE:
                 if attr.localName == 'xmlns':
-                    nss[''] = attr.value
+                    nss[None] = attr.value
                 else:
                     nss[attr.localName] = attr.value
             elif attr.namespaceURI:
@@ -246,9 +255,9 @@ def SplitQName(qname):
         if len(fields) == 1:
             #Note: we could gain a tad more performance by interning 'xmlns'
             if qname == 'xmlns':
-                sName = ('', 'xmlns')
+                sName = (None, 'xmlns')
             else:
-                sName = ('', qname)
+                sName = (None, qname)
         elif len(fields) == 2:
             if fields[0] == 'xmlns':
                 sName = (fields[1], 'xmlns')
@@ -269,6 +278,11 @@ def SeekNss(node, nss=None):
                 nss[child.prefix] = child.namespaceURI
             for attr in child.attributes.values():
                 if attr.namespaceURI == XMLNS_NAMESPACE:
-                    nss[attr.localName] = attr.value
+                    if attr.localName == 'xmlns':
+                        nss[None] = attr.value
+                    else:
+                        nss[attr.localName] = attr.value
+                elif attr.namespaceURI:
+                    nss[attr.prefix] = attr.namespaceURI
             SeekNss(child, nss)
     return nss
